/* There are two modes, one is the split screen mode with the row_list on
 * top, the filtered sc_win, and the cs_win. The user can then hit 'f' to
 * go into full screen mode, where he gives up the cs_win, but gets an
 * unfiltered sc_win 
 */

#include <stdio.h>
#include <sys/types.h>
#include <sys/signal.h>

#include <sys/time.h>
#include <time.h>
#include <termios.h>
#include <curses.h>
#include <fcntl.h>

#ifdef SOLARIS
#include <sys/filio.h>
#endif

#include <errno.h>

#include "pcap.h"
#include "nxm_defines_v0.6b.h"
#include "nxm_external_v0.6b.h"

extern FILE *debug_dev;


int curses_alarm();

/* Used to curses */
WINDOW *connw, *scwin, *cswin, *hlpwin, *cmdwin, *statusw;
char *DASHES="--------------------------------------------------------------------------------";

int curses_fullscreen=0;
int lttyfd=-1;

int top_conn=0;
int cur_conn=0;
int num_conn=0;

init_screen()
{
	struct itimerval value;

	initscr();
	flushinp();
	cbreak();
	noecho();

	if (LINES < 24)
	{
		fprintf(stderr,"Need at least 24 lines\n");
		com_quit();
	}
	do_cool_borders();
	statusw=subwin(stdscr, 8, 23, 2, 57);
	scrollok(statusw, FALSE);
	connw=subwin(stdscr, 8, 56, 2, 0);
	scrollok(connw, FALSE);
	scwin=subwin(stdscr, 11, 80, 11, 0);
	scrollok(scwin, TRUE);
	cswin=subwin(stdscr, 1, 80, 23, 0);
	scrollok(cswin, FALSE);
	cmdwin=subwin(stdscr, 1, 23, 0, 57);
	mvwprintw(cmdwin, 0, 1, "Command:");
	nodelay(cmdwin, TRUE);
	keypad(cmdwin, TRUE);
	refresh();
	/* Setup ocasional maintenance alarms */
	value.it_interval.tv_sec=1;
	value.it_interval.tv_usec=0;
	value.it_value.tv_sec=1;
	value.it_value.tv_usec=0;
	signal(SIGALRM, curses_alarm);
	if (setitimer(ITIMER_REAL, &value, NULL)<0)
	{
		endwin();
		perror("Can't initialize maintenance daemon");
		com_quit();
	}
	return;
}

do_cool_borders()
{
	mvprintw(0, 56, "|");
	mvprintw(1, 0,  DASHES);
	mvprintw(0, 20, "NetXMon V0.6b");
	mvprintw(1, 20, "Active Sessions");
	mvprintw(1, 56, "+");
	mvprintw(2, 56, "|");
	mvprintw(3, 56, "|");
	mvprintw(4, 56, "|");
	mvprintw(5, 56, "|");
	mvprintw(6, 56, "|");
	mvprintw(7, 56, "| Arrow/VI Keys scroll");
	mvprintw(8, 56, "| <ret> selects");
	mvprintw(9, 56, "| [f]ullscreen, [q]uit");
	mvprintw(1, 57, "Status");
	mvprintw(10, 0, DASHES);
	mvprintw(10,32, "Server -> Client");
	mvprintw(22, 0, DASHES);
	mvprintw(22,32, "Client -> Server");
}

redo_scwin(str, len)
char *str;
int len;
{
	char buf[2048];

	if (curses_fullscreen)
	write(1, str, len);
	else
	{
		memset(buf, 0, 2048);
		strncpy(buf, str, len);
		waddstr(scwin, buf);
		wrefresh(scwin);
	}
	return;
}

redo_cswin(buf, n)
char *buf;
int n;
{
	if (curses_fullscreen) return;
	wmove(cswin, 0, 0);
	waddnstr(cswin, buf, n);
	wclrtoeol(cswin);
	wrefresh(cswin);
	return;
}


curses_insert_row(string, cp, row)
char *string;
struct con *cp;  /* Return pcon when this is selected */
int row;
{
	struct con_row *cs, *csp;

	if (row == 0)
	{
  		top_conn=0;
  		cur_conn=0;
	}
	num_conn++;
	do_connw(1);
	return;
}

do_statusw()
{
	char buf[80];
	struct con_row *cs;
	static char old_conn[80], old_status[80];
	static int old_active; 
	int changed=0;

	if (curses_fullscreen) return;
	memset(buf, 0, 80);
	if (current==NULL) strcpy(buf, "No active conn");
	else
	{
		cs=row_list;
		while (cs!=NULL && cs->pcon != current) cs=cs->next;
		if (cs!=NULL) strncpy(buf, cs->name, 31);
	}
	if (strcmp(buf, old_conn))
	{
		strcpy(old_conn, buf);
		mvwprintw(statusw, 0, 0, "Conn: %s", buf);
		if (strlen(buf)<31) wclrtoeol(statusw);
		changed=1;
	}
	if (current!=NULL) 
		strcpy(buf, "*ALIVE*"); 
/* Change this to logging when the time comes */
	else
		strcpy(buf, "*DEAD* ");
	if (strcmp(buf, old_status))
	{
		strcpy(old_status, buf);
		mvwprintw(statusw, 1, 0, "Conn status : %s", buf);
		changed=1;
	}
	if (old_active!=num_conn)
	{
		old_active=num_conn;
		mvwprintw(statusw, 3, 0, "Active conns: %d  ", num_conn);
		changed=1;
	}
	if (changed) wrefresh(statusw);
}

do_connw(num)
int num;
{
	struct con_row *cp;
	int i;
	int did_something=0;
	static int lastnull=0;

	if (curses_fullscreen) return;
	if (row_list==NULL)
	{
		top_conn=0;
		cur_conn=0;
		if (!lastnull)
		{
			werase(connw);
			wrefresh(connw);
			lastnull=1;
		}
		return;
	}
	lastnull=0;
	cp=row_list;
	for(i=0;i<top_conn;i++)
	{
		if (cp->changed) cp->changed=0;
		cp=cp->next;
	}
	while (cp!=NULL)
	{
		if (cp->changed || num)
		{
			did_something=1;
			cp->changed=0;
			if (i==cur_conn)
			{
				wstandout(connw);
				mvwprintw(connw, i-top_conn, 0, "%s%c", cp->name,
				(cp->pcon==current ? '*' : ' '));
				wstandend(connw);
			} else
				mvwprintw(connw, i-top_conn,0,"%s%c",cp->name,
				(cp->pcon==current ? '*' : ' '));
		}
		cp=cp->next;
		i++;
		if (i-top_conn>=8) break;
	}
	for (;i-top_conn<8;i++)
	{
		wmove(connw, i-top_conn, 0);
		wclrtoeol(connw);
	}
	while (cp!=NULL)
	{
		cp->changed=0;
		cp=cp->next;
	}
	if (did_something)
	{
		if (top_conn>0) mvprintw(1, 2, "/\\");
		else mvprintw(1, 2, "--");
		if (top_conn+8<num_conn) mvprintw(10, 2, "\\/");
		else mvprintw(10, 2, "--");
		refresh();
		wrefresh(connw);
		do_statusw();
	}
}

curses_delete_row(row)
int row;
{
	struct con_row *cp, *lp;
	int i;

	num_conn--;
	if (cur_conn>=num_conn) cur_conn=num_conn-1;
	if (top_conn>=num_conn) top_conn=num_conn-1;
	do_connw(1);
	return;
}


curses_redo_row(row, title)
int row;
char *title;
{
	struct con_row *cp;
	int i;

	cp=row_list;
	if (cp==NULL) return;
	for (i=0;i<row;i++) cp=cp->next;
	strncpy(cp->name, title,80);
	cp->changed=1;
	return;
}

void notify_curses()
{
	int c, i;
	struct con_row *cp;
	int redo=0;
	fd_set fdset;
	struct timeval to;
	static struct termios itio, tio;
	char buf[1024];

	if (curses_fullscreen)
	{
		FD_ZERO(&fdset);
		FD_SET(lttyfd, &fdset);
		if (current==NULL)
		{
			while (select(255, &fdset, 0, 0, NULL)<0 && errno==EINTR) ;
			tcsetattr(lttyfd, TCSAFLUSH, &itio);
			close(lttyfd);
			curses_fullscreen=0;
			do_cool_borders();
			wnoutrefresh(scwin);
			wnoutrefresh(cswin);
			wnoutrefresh(cmdwin);
			wnoutrefresh(statusw);
			doupdate();
			keypad(cmdwin,TRUE);
			redo=1;
			return;
		}
		to.tv_sec=0;
		to.tv_usec=0;
		while (select(255, &fdset, 0, 0, &to)>0)
		{
			read(lttyfd, buf, 1);
			if (*buf=='~')
			{
				write(1, "[Escape Mode...]", 16);
				/* Wait a short period of time for the followup */
				to.tv_sec=2;
				to.tv_usec=0;
				FD_ZERO(&fdset);
				FD_SET(lttyfd, &fdset);
				c=select(255, &fdset, 0, 0, &to);
				if (c<0 && errno==EINTR)
				{/*Then the alarm went off, try once more*/
					to.tv_sec=1; /* Don't wait forever */
					to.tv_usec=0;
					FD_ZERO(&fdset);
					FD_SET(lttyfd, &fdset);
					c=select(255, &fdset, 0, 0, &to);
				}
				if (c>0)
				{
					write(1, "\b \b\b \b\b \b\b \b\b \b\b \b\b \b\b \b\b \b\b \b\b \b\b \b\b \b\b \b\b \b\b \b", 48); /* Back up over old stuff */
					while ((c=read(lttyfd, buf, 1))<0 && errno==EINTR) ;
					if (c<=0) return;
					switch (*buf)
					{
						case '.':
							tcsetattr(lttyfd, TCSAFLUSH, &itio);
							close(lttyfd);
							curses_fullscreen=0;
							do_cool_borders();
							wnoutrefresh(scwin);
							wnoutrefresh(cswin);
							wnoutrefresh(cmdwin);
							wnoutrefresh(statusw);
							doupdate();
							keypad(cmdwin,TRUE);
							redo=1;
							break;
						case '?': sprintf(buf, "\n[  =NetXMon Commands=   ]\n[ ~? Print out this message ]\n[ ~. Go back to normal mode ]\n[ ~h Hijacking mode ]\n[ ~k Kill this conn  ]\n");
							  write(1, buf, strlen(buf));
							  break;
						case 'h': 
							  /* Resume data passing to user */
							  /* under developed */
							  /* Will be changed
							   * if we know how to
							   * send a packet */

							  send_packet();
							  break;
						case 'k': /* Kill connection */
							  com_sendrst();
							  break;
						default: 
							  break;
					}
				} else
					write(1, "\b \b\b \b\b \b\b \b\b \b\b \b\b \b\b \b\b \b\b \b\b \b\b \b\b \b\b \b\b \b\b \b", 48); /* Back up over old stuff */
			  	/* under developed */
				/* Will be changed
				 * if we know how to
				 * send a packet */

/*				send_packet(); */
/* Uncomment previous line if you want the ~ sent after a timeout anyway */
			} else
			/* under developed */
			/* Will be changed
			 * if we know how to
			 * send a packet */

			send_packet();
			to.tv_sec=0;
			to.tv_usec=0;
			FD_ZERO(&fdset);
			FD_SET(lttyfd, &fdset);
		}
		return;
	}
	if ((c=mvwgetch(cmdwin, 0, 9))!=ERR)
	switch(c)
	{
		case 'j':
		case 'J':
		case KEY_DOWN:
			if (cur_conn+1<num_conn)
			{
				cur_conn++;
				if (cur_conn-top_conn>=8) top_conn++;
				redo=1;
			}
			break;
		case 'k':
		case 'K':
		case KEY_UP:
			if (cur_conn>0)
			{
				cur_conn--;
				if (cur_conn<top_conn) top_conn--;
				redo=1;
			}
			break;
		case '\r':
		case '\n': 
			cp=row_list;
			for (i=0;i<cur_conn;i++) cp=cp->next;
			if (cp==NULL) break;
			current=cp->pcon;
			werase(scwin);
			werase(cswin);
			wrefresh(scwin);
			wrefresh(cswin);
			dump_bufs(current);
			redo=1;
			break;
		case 'f':
		case 'F':
			if (current==NULL) return;
			curses_fullscreen=1;
			lttyfd=open("/dev/tty", O_RDWR);
			if (lttyfd<0) 
				lttyfd=0;
			endwin();
			/* Put the terminal in CBREAK here */
			tcgetattr(lttyfd, &tio);
			itio=tio;
			tio.c_cc[VMIN] = 1;
			tio.c_cc[VTIME] = 0;
			tio.c_lflag &= ~(ECHO|ICANON|ECHOE|ECHOK|ECHONL|ISIG);
			tcsetattr(lttyfd, TCSAFLUSH, &tio);
			write(lttyfd, "\033[2J\033[0;0H", 10);
			dump_bufs(current);
			break;
		case '': redo=1; break;
		case 'q':
		case 'Q': endwin();
			  com_quit();
		default: break;
	} /* of switch */
	do_connw(redo);
	return;
}

curses_alarm()
{
	gtime++; /* Every second, increment gtime */
	ticks++;
	if (ticks == 60)
	{
		ticks=0;
		delete_hash1();
	}
	signal(SIGALRM, curses_alarm);
	return;
}
